#!/usr/bin/env python
import rospy
import tf
from geometry_msgs.msg import PoseStamped
from math import atan2
import matplotlib.pyplot as plt
from nav_msgs.msg import OccupancyGrid
import numpy as np

# Global variable for storing map data
occupancy_grid = None

def map_callback(msg):
    global occupancy_grid
    occupancy_grid = msg

def log_pose():
    rospy.init_node('log_robot_pose', anonymous=True)

    listener = tf.TransformListener()
    rate = rospy.Rate(10)  # 10 Hz

    # Subscribe to the map topic for Cartographer's map
    rospy.Subscriber("/map", OccupancyGrid, map_callback)

    # Set up the Matplotlib plot
    plt.ion()  # Interactive mode on (for real-time plotting)
    fig, ax = plt.subplots()
    ax.set_xlabel("X Position")
    ax.set_ylabel("Y Position")
    ax.set_title("Robot's Real-Time Position with Cartographer Map")

    # Initial axis limits (adjust if needed)
    ax.set_xlim(-10, 10)  # Set X-axis limit to -10m to 10m
    ax.set_ylim(-10, 10)  # Set Y-axis limit to -10m to 10m

    # Add gridlines
    ax.grid(True, which='both', linestyle='--', linewidth=0.5)  # Add dashed gridlines

    while not rospy.is_shutdown():
        try:
            # Get the transformation from map to base_link
            (trans, rot) = listener.lookupTransform('map', 'base_link', rospy.Time(0))

            # Extract the position
            x, y, z = trans

            # Extract yaw from the quaternion
            yaw = atan2(2.0 * (rot[3] * rot[2] + rot[0] * rot[1]),
                        1.0 - 2.0 * (rot[1] * rot[1] + rot[2] * rot[2]))

            # Log the position
            rospy.loginfo("Position: x={}, y={}, yaw={}".format(x, y, yaw))

            # Clear the previous plot and update with the current position
            ax.clear()
            ax.set_xlabel("X Position")
            ax.set_ylabel("Y Position")
            ax.set_title("Robot's Real-Time Position with Cartographer Map")

            # Dynamically set axis limits based on the robot's position and map size
            if occupancy_grid:
                map_width = occupancy_grid.info.width * occupancy_grid.info.resolution
                map_height = occupancy_grid.info.height * occupancy_grid.info.resolution

                # Optionally, expand the plot limits based on the map size
                ax.set_xlim(-map_width/2, map_width/2)
                ax.set_ylim(-map_height/2, map_height/2)

            # Plot the current position (robot's position)
            ax.plot(x, y, marker='o', color='b', label='Current Position')

            # Add the robot's coordinates as text above the point
            ax.text(x, y + 0.2, '({:.2f}, {:.2f})'.format(x, y), color='red', fontsize=10, ha='center')

            # If occupancy grid data is available, plot the map
            if occupancy_grid:
                # Convert the map data into a 2D array
                map_data = np.array(occupancy_grid.data).reshape((occupancy_grid.info.height, occupancy_grid.info.width))

                # Plot the occupancy grid (Cartographer map)
                ax.imshow(map_data, cmap="gray", extent=[-map_width/2, map_width/2, -map_height/2, map_height/2], origin="lower", alpha=0.5)

            # Add gridlines and re-enable them
            ax.grid(True, which='both', linestyle='--', linewidth=0.5)

            # Add a legend for the current position marker
            ax.legend()

            plt.draw()
            plt.pause(0.1)  # Pause for a short period to allow for real-time update

        except (tf.LookupException, tf.ConnectivityException, tf.ExtrapolationException) as e:
            rospy.logwarn("Transform error: {}".format(e))

        rate.sleep()

    # Keep the plot open after the loop finishes
    plt.ioff()
    plt.show()

if __name__ == "__main__":
    try:
        log_pose()
    except rospy.ROSInterruptException:
        pass

