#!/usr/bin/env python
import rospy
import tf
from geometry_msgs.msg import PoseStamped
from math import atan2
import matplotlib.pyplot as plt
from matplotlib import image as mpimg
import numpy as np
from PIL import Image  # Using PIL to read image dimensions
import yaml

# Path to the pre-generated map file (.pgm)
map_file = "/home/sepiot2/catkin_ws/maps/maps.pgm"  # Update with your actual map file path
map_yaml_file = "/home/sepiot2/catkin_ws/maps/maps.yaml"  # Update to your actual map.yaml path

# Function to load map.yaml and extract parameters
def load_map_yaml(yaml_file):
    with open(yaml_file, 'r') as file:
        map_info = yaml.load(file, Loader=yaml.SafeLoader)  # Use SafeLoader
    return map_info

def log_pose():
    rospy.init_node('log_robot_pose', anonymous=True)

    listener = tf.TransformListener()
    rate = rospy.Rate(10)  # 10 Hz

    # Set up the Matplotlib plot
    plt.ion()  # Interactive mode on (for real-time plotting)
    fig, ax = plt.subplots(figsize=(10, 10))  # Adjust figure size to match the map
    ax.set_xlabel("X Position")
    ax.set_ylabel("Y Position")
    ax.set_title("Robot's Real-Time Position with Map")

    # Load the map info from the YAML file
    map_info = load_map_yaml(map_yaml_file)

    # Map parameters from the YAML file
    resolution = map_info['resolution']
    origin_x, origin_y, _ = map_info['origin']

    # Get the dimensions of the map image from the .pgm file
    img = Image.open(map_file)  # Open the map image
    map_width, map_height = img.size  # Get the width and height of the image

    # Load the pre-generated map (.pgm file)
    img = mpimg.imread(map_file)

    # Calculate the map's extent based on resolution and origin
    xmin = origin_x
    xmax = origin_x + map_width * resolution
    ymin = origin_y
    ymax = origin_y + map_height * resolution

    # Plot the map with correct extent
    ax.imshow(img, cmap='gray', extent=[xmin, xmax, ymin, ymax], origin='lower', alpha=0.5)

    # Set aspect ratio to be equal
    ax.set_aspect('auto')  # You can try 'equal' if you want no stretching

    # Add gridlines
    ax.grid(True, which='both', linestyle='--', linewidth=0.5)

    while not rospy.is_shutdown():
        try:
            # Get the transformation from map to base_link
            (trans, rot) = listener.lookupTransform('map', 'base_link', rospy.Time(0))

            # Extract the position
            x, y, z = trans

            # Extract yaw from the quaternion
            yaw = atan2(2.0 * (rot[3] * rot[2] + rot[0] * rot[1]),
                        1.0 - 2.0 * (rot[1] * rot[1] + rot[2] * rot[2]))

            # Log the position
            rospy.loginfo("Position: x={}, y={}, yaw={}".format(x, y, yaw))

            # Clear the previous plot and update with the current position
            ax.clear()
            ax.set_xlabel("X Position")
            ax.set_ylabel("Y Position")
            ax.set_title("Robot's Real-Time Position with Map")

            # Replot the map in the background
            ax.imshow(img, cmap='gray', extent=[xmin, xmax, ymin, ymax], origin='lower', alpha=0.5)

            # Plot the current position (robot's position)
            ax.plot(x, y, marker='o', color='b', label='Current Position')

            # Add the robot's coordinates as text above the point
            ax.text(x, y + 0.2, '({:.2f}, {:.2f})'.format(x, y), color='red', fontsize=10, ha='center')

            # Add gridlines and re-enable them
            ax.grid(True, which='both', linestyle='--', linewidth=0.5)

            # Add a legend for the current position marker
            ax.legend()

            plt.draw()
            plt.pause(0.1)  # Pause for a short period to allow for real-time update

        except (tf.LookupException, tf.ConnectivityException, tf.ExtrapolationException) as e:
            rospy.logwarn("Transform error: {}".format(e))

        rate.sleep()

    # Keep the plot open after the loop finishes
    plt.ioff()
    plt.show()

if __name__ == "__main__":
    try:
        log_pose()
    except rospy.ROSInterruptException:
        pass

